#!/usr/bin/env python

""" This module is a port of Nigel Stewart's ZPR C module
 Which is available under the LGPL, thus this code is also LGPL
 Ported by Thomas Dickerson """

#
#  Copyright Thomas Dickerson, 2012.
#  This file is part of PyZPR.
#
#  PyZPR is free software: you can redistribute it and/or modify it
#  under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  PyZPR is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  long with PyZPR.  If not, see <http://www.gnu.org/licenses/>.
#

import sys,math

try:
	from numpy import array, float32, matrix, linalg, empty, uint32
except:
	print "Oops, please install numpy on this machine!"
	sys.exit(1)

try:
	from OpenGL.GLUT import GLUT_UP, GLUT_DOWN, GLUT_MIDDLE_BUTTON, GLUT_RIGHT_BUTTON, GLUT_LEFT_BUTTON, GLUT_WINDOW_HEIGHT
	from OpenGL.GLUT import glutGet, glutPostRedisplay, glutReshapeFunc, glutMouseFunc, glutMotionFunc
	
	from OpenGL.GLU import gluPickMatrix
	
	from OpenGL.GL import GL_PROJECTION, GL_MODELVIEW, GL_MODELVIEW_MATRIX, GL_SELECT, GL_PROJECTION
	from OpenGL.GL import GL_VIEWPORT, GL_PROJECTION_MATRIX, GL_RENDER
	from OpenGL.GL import glViewport, glMatrixMode, glLoadIdentity, glOrtho, glGetDoublev, glSelectBuffer, glRenderMode, glInitNames
	from OpenGL.GL import glGetIntegerv, glMultMatrixd, glPushMatrix, glPopMatrix, glTranslatef, glRotatef, glScalef
except:
	print "Oops, please install PyOpenGL on this machine!"
	sys.exit(1)

left = right = 0
top = 1
bottom = -top
far = 10
near = -far
angle = 0

ref_pt = array((0.,0.,0.,0.))
nref_pt = array((-0.,-0.,-0.,-0.))

mat = matrix([[1.,0.,0.,0.],[0.,1.,0.,0.],[0.,0.,1.,0.],[0.,0.,0.,1.]])
inv_mat = linalg.inv(mat)

mouse_states = {GLUT_LEFT_BUTTON : GLUT_UP, GLUT_MIDDLE_BUTTON : GLUT_UP, GLUT_RIGHT_BUTTON : GLUT_UP}
mouse_pos = (0,0)
drag_pos = (0., 0., 0.)

pick_cb = None
sel_cb = None
SELECT_BUFFER_SIZE = 1024

def handleResize(width, height):
	
	global left, right, bottom, top, near, far
	
	glViewport(0, 0, width, height)
	
	
	if height <= 0:
		height = 1
	ratio = float(width)/height
	top = 1.
	bottom = -1.
	left = -ratio
	right = ratio
		
	glMatrixMode(GL_PROJECTION)
	glLoadIdentity()
		
	glOrtho(left, right, bottom, top, near, far)
	
	glMatrixMode(GL_MODELVIEW)
	
def getMatrix():
	global mat, inv_mat
	mat = matrix(glGetDoublev(GL_MODELVIEW_MATRIX))
	inv_mat = linalg.inv(mat)

def pick(x, y, dX, dY):
	global pick_cb, sel_cb, SELECT_BUFFER_SIZE
	buffer = empty(SELECT_BUFFER_SIZE, dtype=uint32)
	
	min = -1
	minZ = -1
	
	glSelectBuffer(SELECT_BUFFER_SIZE,buffer)
	glRenderMode(GL_SELECT)
	glInitNames()
	glMatrixMode(GL_PROJECTION)
	glPushMatrix();
	viewport = array(glGetIntegerv(GL_VIEWPORT))
	projection = matrix(glGetDoublev(GL_PROJECTION_MATRIX))
	glLoadIdentity()
	gluPickMatrix(x,y,dX,dY,viewport)
	glMultMatrixd(projection)
	
	glMatrixMode(GL_MODELVIEW)
	
	if callable(sel_cb):
		sel_cb()
		
	hit_list = glRenderMode(GL_RENDER)
	hits = len(hit_list)
	
	if hits:
		j = 0
		for i in range(hits):
			if buffer[j+1] < minZ:
				if buffer[j]:
					min = -1
				else:
					min = buffer[j+2+buffer[j]]
				minZ = buffer[j+1]
			j += buffer[j] + 3
	glMatrixMode(GL_PROJECTION)
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW)
	
	if callable(pick_cb):
		pick_cb(min)
		
def calc_pos(x, y, viewport):
	global left, right, top, bottom, near, far
	return (left + float(x-viewport[0])/float(viewport[2]) * (right - left),
			top + float(y-viewport[1])/float(viewport[3]) * (bottom - top),
			near)
	
	

def handleMouseEvent(button, state, mouseX, mouseY):
	global mouse_states, mouse_pos, drag_pos
	
	if state == GLUT_DOWN:
		pick(mouseX, glutGet(GLUT_WINDOW_HEIGHT)-1-mouseY,3,3)
	
	mouse_states[button] = state
	mouse_pos = (mouseX,mouseY)
	
	viewport = array(glGetIntegerv(GL_VIEWPORT))
	drag_pos = calc_pos(mouseX, mouseY, viewport)
	glutPostRedisplay()
	
def handleMotionEvent(mouseX, mouseY):
	global mouse_pos, mouse_states, mat, inv_mat, drag_pos
	changed = False
	
	viewport = array(glGetIntegerv(GL_VIEWPORT))
	
	oldX, oldY = mouse_pos
	dx = mouseX - oldX
	dy = mouseY - oldY
	if dx or dy:
		changed = True
		if mouse_states[GLUT_MIDDLE_BUTTON] == GLUT_DOWN or (mouse_states[GLUT_LEFT_BUTTON] ==  mouse_states[GLUT_RIGHT_BUTTON] == GLUT_DOWN):
			s = math.exp(0.01 * dy)
			glTranslatef(*ref_pt[:3])
			glScalef(s,s,s)
			glTranslatef(*nref_pt[:3])
		elif mouse_states[GLUT_LEFT_BUTTON] == GLUT_DOWN:
			axis = (dy,dx,0.)
			angle = 180.*linalg.norm(axis)/(1+viewport[2])
			baxis = [sum(inv_mat[i,j] * x for j,x in enumerate(axis)) for i in range(3)]
			glTranslatef(*ref_pt[:3])
			glRotatef(angle, *baxis)
			glTranslatef(*nref_pt[:3])
		elif mouse_states[GLUT_RIGHT_BUTTON] == GLUT_DOWN:
			pos = calc_pos(mouseX, mouseY, viewport)
			
			glLoadIdentity()
			glTranslatef(*(pos[i] - drag_pos[i] for i in range(3)))
			glMultMatrixd(mat)
			
			drag_pos = pos
			
	mouse_pos = (mouseX, mouseY)
	if changed:
		getMatrix()
		glutPostRedisplay()

def set_sel_cb(scb=None):
	global sel_cb
	sel_cb = scb
	
def set_pick_cb(pcb=None):
	global pick_cb
	pick_cb = pcb

def set_ref_pt(rpt=(0.,0.,0.)):
	global ref_pt,nref_pt
	
	ref_pt = rpt
	nref_pt = array([-i for i in ref_pt])


def init(rpt=(0.,0.,0.),scb=None,pcb=None):
	getMatrix()
	glutReshapeFunc(handleResize)
	glutMouseFunc(handleMouseEvent)
	glutMotionFunc(handleMotionEvent)
	
	set_sel_cb(scb=scb)
	set_pick_cb(pcb=pcb)
	set_ref_pt(rpt)